"use strict";


const system = server.registerSystem(0, 0);


system.initialize = function () {
    Object.defineProperty(this, "query", {
        value: this.registerQuery(),
        writable: false,
        configurable: false
    })
};

system.update = (function () {
    let radius = 0;
    const flippedChunks = new Set;
    const chunksDone = new Set;

    function pushCallStack(thisArg, context) {
        if (context.callCallStack == null) {
            const node = context.instructionSequence.next();
            if (node.done) {
                const msg = system.createEventData("minecraft:display_chat_event");
                msg.data.message = "unexpected processor emulation shutdown";
                system.broadcastEvent("minecraft:display_chat_event", msg);
            } else {
                context.callCallStack = node.value
            }
        }
        const result = context.callCallStack.apply(thisArg);
        if (result != true) {
            context.callCallStack = null
        }
    }

    function* generateFlipSequence(chunks, popCallStack) {
        for (let [chunkX, chunkZ] of chunks) {
            let context;
            const identifier = `${chunkX},${chunkZ}`;
            yield (function () {
                if (chunksDone.has(identifier)) {
                    context = true;
                    return;
                }
                context = null;
                flippedChunks.add(identifier);
                const [x, z] = [chunkX * 16, chunkZ * 16];
                this.executeCommand(`/testforblock ${x} 319 ${z} chain_command_block 0`, (result) => {
                    context = true;
                    if (chunksDone.has(identifier)) {
                        return;
                    }
                    const { matches } = result.data;
                    if (matches == false) {
                        flippedChunks.add(identifier);

                        const command = this.createEventData("minecraft:execute_command");
                        command.data.command = `/clone ${x} 128 ${z} ${x + 15} 191 ${z + 15} ${x} 256 ${z} masked move`;
                        this.broadcastEvent("minecraft:execute_command", command);
                        let y = 127;
                        do {
                            command.data.command = `/clone ${x} ${y} ${z} ${x + 15} ${y} ${z + 15} ${x} ${255 - y} ${z} masked move`;
                            this.broadcastEvent("minecraft:execute_command", command);
                            const offset = (255 - y * 2);
                            if (offset != 0) {
                                command.data.command = `/execute @e[ x=${x},y=${y},z=${z},dx=15,dy=0,dz=15 ] ~ ~ ~ teleport @s ~ ~${offset} ~`;
                                this.broadcastEvent("minecraft:execute_command", command)
                            }
                            y--
                        } while (y != -1);
                        for (let i = 256; i != 320; i++) {
                            command.data.command = `/clone ${x} ${i} ${z} ${x + 15} ${i} ${z + 15} ${x} ${383 - i} ${z} masked move`;
                            this.broadcastEvent("minecraft:execute_command", command)
                        }
                        do {
                            command.data.command = `/clone ${x} ${y} ${z} ${x + 15} ${y} ${z + 15} ${x} ${255 - y} ${z} masked move`;
                            this.broadcastEvent("minecraft:execute_command", command);
                            const offset = (255 - y * 2);
                            command.data.command = `/execute @e[ x=${x},y=${y},z=${z},dx=15,dy=0,dz=15 ] ~ ~ ~ teleport @s ~ ~${(offset == 0) ? empty : offset} ~`;
                            this.broadcastEvent("minecraft:execute_command", command);
                            y--
                        } while (y != -65);
                        command.data.command = `/setblock ${x} 319 ${z} chain_command_block 0`;
                        this.broadcastEvent("minecraft:execute_command", command);
                        context = false
                    } else if (matches == true) {
                        chunksDone.add(identifier);
                        flippedChunks.delete(identifier)
                    }
                })
            });
            yield (function () {
                if (context == null) {
                    return true;
                }
                if (context) {
                    return;
                }
                this.executeCommand(`/testforblock ${chunkX * 16} 319 ${chunkZ * 16} chain_command_block 0`, (result) => {
                    flippedChunks.delete(identifier);
                    if (result.data.matches == true) {
                        chunksDone.add(identifier)
                    }
                })
            });
        }
        yield popCallStack;
    }

    function* generateChunkHandlersSequence(x, z, popCallStack) {
        let context;
        yield (function () {
            let [chunkX, chunkZ] = [x, z];
            let step = 1;
            let stepSign = 1;
            const chunks = new Array;
            chunks.push([chunkX, chunkZ]);
            while (step != radius * 2 + 1) {
                const [boundX, boundZ] = [(chunkX + step * stepSign), (chunkZ + step * stepSign)];
                while (chunkZ != boundZ) {
                    chunkZ += stepSign;
                    chunks.push([chunkX, chunkZ])
                }
                while (chunkX != boundX) {
                    chunkX += stepSign;
                    chunks.push([chunkX, chunkZ])
                }
                stepSign = 0 - stepSign;
                step++
            }
            const boundZ = chunkZ + step * stepSign - 1;
            while (chunkZ != boundZ) {
                chunkZ += stepSign;
                chunks.push([chunkX, chunkZ])
            }
            context = chunks
        });
        yield (function () {
            const ctx = {
                callCallStack: null,
                closed: false,
                instructionSequence: generateFlipSequence(context, () => {
                    ctx.closed = true
                })
            };
            context = ctx
        });
        yield (function () {
            if (context.closed) {
                return;
            }
            pushCallStack(this, context);
            return true;
        });
        yield popCallStack;
    }
    const positionsDone = new Set;
    function* generatePositionHandlerLoopSequence(positions, popCallStack) {
        for (let position of positions) {
            const [x, z] = [Math.floor(position.data.x / 16), Math.floor(position.data.z / 16)];
            const positionIdentifier = `${x},${z}`;
            if (positionsDone.has(positionIdentifier)) {
                continue
            }
            positionsDone.add(positionIdentifier);

            let context;
            yield (function () {
                context = {
                    callCallStack: null,
                    closed: false,
                    instructionSequence: generateChunkHandlersSequence(x, z, () => {
                        context.closed = true
                    })
                }
            });
            yield (function () {
                if (context.closed) {
                    return;
                }
                pushCallStack(this, context);
                return true;
            });
        }
        yield popCallStack;
    }
    function* generateMainSequence(popCallStack) {
        let context;

        let player = null;
        yield (function () {
            this.executeCommand("/testfor @a[ rm=0 ]", (result) => {
                const overworldPlayers = new Set(result.data.victim);
                const entities = this.getEntitiesFromQuery(this.query);
                const positions = new Set;
                for (let entity of entities) {
                    if (entity.__identifier__ != "minecraft:player" || !(this.hasComponent(entity, "minecraft:nameable")) || !(overworldPlayers.has(this.getComponent(entity, "minecraft:nameable").data.name))) {
                        continue
                    }
                    positions.add(this.getComponent(entity, "minecraft:position"));
                    if (radius == 0 && player == null) {
                        player = entity
                    }
                }
                context = positions
            });
            context = null
        });
        yield (function () {
            if (context == null) { // wait for command result
                return true;
            }

            if (radius == 0 && player != null) {
                radius = this.getComponent(player, "minecraft:tick_world").data.radius
            }

            const ctx = {
                callCallStack: null,
                // notBreaking: true,
                closed: false,
                instructionSequence: generatePositionHandlerLoopSequence(context, () => {
                    ctx.closed = true
                }/*, () => {
                    ctx.notBreaking = false
                }*/)
            };
            context = ctx
        });
        yield (function () {
            if (context.closed) { // current sequence will be call stack top
                return; // continue current sequence
            }
            pushCallStack(this, context);
            return true; // check call stack again
        });
        yield popCallStack;
    }


    const context = {
        callCallStack: null,
        instructionSequence: generateMainSequence((function () {
            const loop = function () {
                context.instructionSequence = generateMainSequence(loop)
            };
            return loop;
        })())
    };


    let skippingTick = true;
    return (function () {
        if (skippingTick) {
            skippingTick = false;
            return;
        }
        skippingTick = true;

        pushCallStack(this, context)
    });
})();
